import logging
import itertools
import numpy as np
import pandas as pd
from pymoo.indicators.hv import HV
from pymoo.config import Config
from pyaugmecon.logs import Logs
from pyaugmecon.model import Model
from pyaugmecon.helper import Helper, Timer
from pyaugmecon.options import Options
from pyaugmecon.queue_handler import QueueHandler
from pyaugmecon.process_handler import ProcessHandler


class PyAugmecon(object):
    def __init__(self, model, opts, solver_opts={}):
        self.opts = Options(opts, solver_opts)
        self.logs = Logs(self.opts)
        self.logger = logging.getLogger(self.opts.log_name)
        self.logger.setLevel(logging.INFO)
        self.opts.log()
        self.model = Model(model, self.opts)
        self.opts.check(self.model.n_obj)
        Config.warnings["not_compiled"] = False  # suppress pymoo warnings

    def find_solutions(self):
        self.model.progress.set_message("finding solutions")

        grid_range = range(self.opts.gp)
        indices = [tuple([n for n in grid_range]) for _ in self.model.iter_obj2]
        self.cp = list(itertools.product(*indices))
        self.cp = [i[::-1] for i in self.cp]

        self.model.pickle()
        self.queues = QueueHandler(self.cp, self.opts)
        self.queues.split_work()
        self.procs = ProcessHandler(self.opts, self.model, self.queues)

        self.procs.start()
        self.unprocesssed_sols = self.queues.get_result()
        self.procs.join()
        self.model.clean()

    def process_solutions(self):
        def convert_obj_goal(sols):
            return np.array(sols) * self.model.obj_goal

        def keep_undominated(pts):
            pts = np.array(pts)
            undominated = np.ones(pts.shape[0], dtype=bool)
            for i, c in enumerate(pts):
                if undominated[i]:
                    undominated[undominated] = np.any(pts[undominated] > c, axis=1)
                    undominated[i] = True

            return pts[undominated, :]

        # Remove duplicate solutions
        self.sols = list(set(tuple(self.unprocesssed_sols)))
        self.sols = [list(i) for i in self.sols]
        self.num_sols = len(self.sols)

        # Remove duplicate solutions due to numerical issues by rounding
        self.unique_sols = [tuple(round(sol, self.opts.round) for sol in item) for item in self.sols]
        self.unique_sols = list(set(tuple(self.unique_sols)))
        self.unique_sols = [list(i) for i in self.unique_sols]
        self.num_unique_sols = len(self.unique_sols)

        # Remove dominated solutions
        self.unique_pareto_sols = keep_undominated(self.unique_sols)
        self.num_unique_pareto_sols = len(self.unique_pareto_sols)

        # Multiply by -1 if original objective was minimization
        self.model.payoff = convert_obj_goal(self.model.payoff)
        self.sols = convert_obj_goal(self.sols)
        self.unique_sols = convert_obj_goal(self.unique_sols)
        self.unique_pareto_sols = convert_obj_goal(self.unique_pareto_sols)

    def output_excel(self):
        writer = pd.ExcelWriter(f"{self.logs.logdir}{self.opts.log_name}.xlsx")
        pd.DataFrame(self.model.e).to_excel(writer, "e_points")
        pd.DataFrame(self.model.payoff).to_excel(writer, "payoff_table")
        pd.DataFrame(self.sols).to_excel(writer, "sols")
        pd.DataFrame(self.unique_sols).to_excel(writer, "unique_sols")
        pd.DataFrame(self.unique_pareto_sols).to_excel(writer, "unique_pareto_sols")
        writer.close()

    def get_hv_indicator(self):
        ref = np.diag(self.model.payoff)
        ind = HV(ref_point=ref)
        self.hv_indicator = ind(self.unique_pareto_sols)

    def solve(self):
        self.runtime = Timer()
        self.model.min_to_max()
        self.model.construct_payoff()
        self.model.find_obj_range()
        self.model.convert_prob()
        self.find_solutions()
        self.process_solutions()
        self.get_hv_indicator()
        if self.opts.output_excel:
            self.output_excel()

        self.runtime = round(self.runtime.get(), 2)
        Helper.clear_line()
        print(
            f"Solved {self.model.models_solved.value()} models for "
            f"{self.num_unique_pareto_sols} unique Pareto solutions in "
            f"{self.runtime} seconds"
        )

        self.logger.info(Helper.separator())
        self.logger.info(f"Runtime: {self.runtime} seconds")
        self.logger.info(f"Models solved: {self.model.models_solved.value()}")
        self.logger.info(f"Infeasibilities: {self.model.infeasibilities.value()}")
        self.logger.info(f"Solutions: {self.num_sols}")
        self.logger.info(f"Unique solutions: {self.num_unique_sols}")
        self.logger.info(f"Unique Pareto solutions: {self.num_unique_pareto_sols}")
        self.logger.info(f"Hypervolume indicator: {self.hv_indicator}")



import os
import logging
from pathlib import Path
from pyaugmecon.options import Options

logging.getLogger("pyomo.core").setLevel(logging.ERROR)


class Logs(object):
    def __init__(self, opts: Options):
        self.opts = opts

        if not os.path.exists(self.opts.logdir):
            os.makedirs(self.opts.logdir)

        self.logdir = f"{Path().absolute()}/{self.opts.logdir}/"
        self.logfile = f"{self.logdir}{self.opts.log_name}.log"
        self.logger = logging.getLogger(opts.log_name)
        self.handler = logging.FileHandler(self.logfile)
        self.formatter = logging.Formatter("[%(asctime)s] %(message)s")
        self.handler.setFormatter(self.formatter)
        self.logger.addHandler(self.handler)
        self.handler.setLevel(logging.INFO)



import itertools
from multiprocessing import Manager
from pyaugmecon.options import Options


class Flag(object):
    def __init__(self, opts: Options):
        self.opts = opts

        if self.opts.shared_flag:
            self.flag = Manager().dict()
        else:
            self.flag = {}

    def set(self, flag_range, value, iter):
        indices = [tuple([n for n in flag_range(o)]) for o in iter]
        iter = list(itertools.product(*indices))
        tmp_flag = {}

        for gp in iter:
            tmp_flag[gp] = value

        self.flag.update(tmp_flag)

    def get(self, i):
        return self.flag.get(i, 0)



import time
import logging
from threading import Thread
from pyaugmecon.flag import Flag
from pyaugmecon.model import Model
from pyaugmecon.helper import Timer
from pyaugmecon.options import Options
from pyaugmecon.queue_handler import QueueHandler
from pyaugmecon.solver_process import SolverProcess


class ProcessHandler(object):
    def __init__(self, opts: Options, model: Model, queues: QueueHandler):
        self.opts = opts
        self.model = model
        self.queues = queues
        self.logger = logging.getLogger(opts.log_name)
        self.flag = Flag(self.opts)

        if self.opts.process_timeout:
            self.timeout = Thread(target=self.check_timeout)

        self.procs = [
            SolverProcess(p_num, self.opts, self.model, self.queues, self.flag)
            for p_num in range(self.queues.proc_count)
        ]

    def start(self):
        self.runtime = Timer()
        self.logger.info(f"Starting {self.queues.proc_count} worker process(es)")

        for p in self.procs:
            p.start()

        if self.opts.process_timeout:
            self.timeout.start()

    def check_timeout(self):
        while self.runtime.get() <= self.opts.process_timeout:
            if not any(p.is_alive() for p in self.procs):
                break
            time.sleep(0.5)
        else:
            self.logger.info("Timed out, gracefully stopping all worker proces(es)")
            self.queues.empty_job_qs()

    def join(self):
        self.logger.info(f"Joining {self.queues.proc_count} worker process(es)")

        if self.opts.process_timeout:
            self.timeout.join()

        for p in self.procs:
            p.join()



import sys
from pathlib import Path
from setuptools import setup

CURRENT_DIR = Path(__file__).parent
sys.path.insert(0, str(CURRENT_DIR))  # for setuptools.build_meta


def get_long_description() -> str:
    return (
        (CURRENT_DIR / "README.md").read_text(encoding="utf8")
        + "\n\n"
        + (CURRENT_DIR / "CHANGELOG.md").read_text(encoding="utf8")
    )


setup(
    name="pyaugmecon",
    version="0.1.9",
    author="Wouter Bles",
    author_email="whbles@gmail.com",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    keywords="python, pyomo, optimization, multi-objective-optimization, augmecon",
    url="https://github.com/wouterbles/pyaugmecon",
    project_urls={"Changelog": "https://github.com/wouterbles/pyaugmecon/blob/main/CHANGELOG.md"},
    license="MIT",
    packages=["pyaugmecon"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Education",
    ],
    test_suite="tests",
    python_requires=">=3.8",
    install_requires=[
        "pyomo>=6.4,<6.5",
        "numpy>=1.2,<1.25",
        "pandas>=1.2,<1.6",
        "cloudpickle>=2.0,<2.3",
        "pymoo>=0.6,<0.7",
        "openpyxl>=3.0,<3.1",
    ],
)
